@@ -11,260 +11,10 @@ import sys
 
 import HTSeq
 from HTSeq._HTSeq import *
+from HTSeq.utils import FileOrSequence
+from HTSeq.features import *
 from HTSeq._version import __version__
 
-#########################
-# Utils
-#########################
-
-
-class FileOrSequence(object):
-    """ The construcutor takes one argument, which may either be a string,
-    which is interpreted as a file name (possibly with path), or a
-    connection, by which we mean a text file opened for reading, or
-    any other object that can provide an iterator over strings
-    (lines of the file).
-
-    The advantage of passing a file name instead of an already opened file
-    is that if an iterator is requested several times, the file will be
-    re-opened each time. If the file is already open, its lines can be read
-    only once, and then, the iterator stays exhausted.
-
-    Furthermore, if a file name is passed that end in ".gz" or ".gzip"
-    (case insensitive), it is transparently gunzipped.
-    """
-
-    def __init__(self, filename_or_sequence):
-        self.fos = filename_or_sequence
-        self.line_no = None
-
-    def __iter__(self):
-        self.line_no = 1
-        if isinstance(self.fos, str):
-            if self.fos.lower().endswith((".gz", ".gzip")):
-                lines = gzip.open(self.fos, 'rt')
-            else:
-                lines = open(self.fos)
-        else:
-            lines = self.fos
-        for line in lines:
-            yield line
-            self.line_no += 1
-        if isinstance(self.fos, str):
-            lines.close()
-        self.line_no = None
-
-    def __repr__(self):
-        if isinstance(self.fos, str):
-            return "<%s object, connected to file name '%s'>" % (
-                self.__class__.__name__, self.fos)
-        else:
-            return "<%s object, connected to %s >" % (
-                self.__class__.__name__, repr(self.fos))
-
-    def get_line_number_string(self):
-        if self.line_no is None:
-            if isinstance(self.fos, str):
-                return "file %s closed" % self.fos
-            else:
-                return "file closed"
-        if isinstance(self.fos, str):
-            return "line %d of file %s" % (self.line_no, self.fos)
-        else:
-            return "line %d" % self.line_no
-
-
-#########################
-# Features
-#########################
-
-class GenomicFeature(object):
-    """A genomic feature, i.e., an interval on a genome with metadata.
-
-    At minimum, the following information should be provided by slots:
-
-      name: a string identifying the feature (e.g., a gene symbol)
-      type: a string giving the feature type (e.g., "gene", "exon")
-      iv: a GenomicInterval object specifying the feature locus
-    """
-
-    def __init__(self, name, type_, interval):
-        self.name = name
-        self.type = sys.intern(type_)
-        self.iv = interval
-
-    def __repr__(self):
-        return "<%s: %s '%s' at %s: %d -> %d (strand '%s')>" % \
-            (self.__class__.__name__, self.type, self.name,
-             self.iv.chrom, self.iv.start_d, self.iv.end_d, self.iv.strand)
-
-    def __eq__(self, other):
-        if not isinstance(other, GenomicFeature):
-            return False
-        return self.name == other.name and self.type == other.type and \
-            self.iv == other.iv
-
-    def __neq__(self, other):
-        if not isinstance(other, GenomicFeature):
-            return True
-        return not self.__eq__(other)
-
-    def get_gff_line(self, with_equal_sign=False):
-        try:
-            source = self.source
-        except AttributeError:
-            source = "."
-        try:
-            score = self.score
-        except AttributeError:
-            score = "."
-        try:
-            frame = self.frame
-        except AttributeError:
-            frame = "."
-        try:
-            attr = self.attr
-        except AttributeError:
-            attr = {'ID': self.name}
-        if with_equal_sign:
-            sep = "="
-        else:
-            sep = " "
-        attr_str = '; '.join(
-            ['%s%s\"%s\"' % (ak, sep, attr[ak]) for ak in attr])
-        return "\t".join(str(a) for a in (self.iv.chrom, source,
-                         self.type, self.iv.start + 1, self.iv.end, score,
-                         self.iv.strand, frame, attr_str)) + "\n"
-
-
-_re_attr_main = re.compile("\s*([^\s\=]+)[\s=]+(.*)")
-_re_attr_empty = re.compile("^\s*$")
-
-
-def parse_GFF_attribute_string(attrStr, extra_return_first_value=False):
-    """Parses a GFF attribute string and returns it as a dictionary.
-
-    If 'extra_return_first_value' is set, a pair is returned: the dictionary
-    and the value of the first attribute. This might be useful if this is the
-    ID.
-    """
-    if attrStr.endswith("\n"):
-        attrStr = attrStr[:-1]
-    d = {}
-    first_val = "_unnamed_"
-    for (i, attr) in zip(
-            itertools.count(),
-            _HTSeq.quotesafe_split(attrStr.encode())):
-        attr = attr.decode()
-        if _re_attr_empty.match(attr):
-            continue
-        if attr.count('"') not in (0, 2):
-            raise ValueError(
-                "The attribute string seems to contain mismatched quotes.")
-        mo = _re_attr_main.match(attr)
-        if not mo:
-            raise ValueError("Failure parsing GFF attribute line")
-        val = mo.group(2)
-        if val.startswith('"') and val.endswith('"'):
-            val = val[1:-1]
-        d[sys.intern(mo.group(1))] = sys.intern(val)
-        if extra_return_first_value and i == 0:
-            first_val = val
-    if extra_return_first_value:
-        return (d, first_val)
-    else:
-        return d
-
-
-_re_gff_meta_comment = re.compile("##\s*(\S+)\s+(\S*)")
-
-
-class GFF_Reader(FileOrSequence):
-    """Parse a GFF file
-
-    Pass the constructor either a file name or an iterator of lines of a
-    GFF files. If a file name is specified, it may refer to a gzip compressed
-    file.
-
-    Iterating over the object then yields GenomicFeature objects.
-    """
-
-    def __init__(self, filename_or_sequence, end_included=True):
-        FileOrSequence.__init__(self, filename_or_sequence)
-        self.end_included = end_included
-        self.metadata = {}
-
-    def __iter__(self):
-        for line in FileOrSequence.__iter__(self):
-            if isinstance(line, bytes):
-                line = line.decode()
-            if line == "\n":
-                continue
-            if line.startswith('#'):
-                if line.startswith("##"):
-                    mo = _re_gff_meta_comment.match(line)
-                    if mo:
-                        self.metadata[mo.group(1)] = mo.group(2)
-                continue
-            (seqname, source, feature, start, end, score,
-             strand, frame, attributeStr) = line.split("\t", 8)
-            (attr, name) = parse_GFF_attribute_string(attributeStr, True)
-            if self.end_included:
-                iv = GenomicInterval(
-                        seqname,
-                        int(start) - 1, int(end),
-                        strand)
-            else:
-                iv = GenomicInterval(
-                        seqname,
-                        int(start) - 1, int(end) - 1,
-                        strand)
-            f = GenomicFeature(name, feature, iv)
-            if score != ".":
-                score = float(score)
-            if frame != ".":
-                frame = int(frame)
-            f.source = source
-            f.score = score
-            f.frame = frame
-            f.attr = attr
-            yield f
-
-
-def make_feature_dict(feature_sequence):
-    """A feature dict is a convenient way to organize a sequence of Feature
-    object (which you have got, e.g., from parse_GFF).
-
-    The function returns a dict with all the feature types as keys. Each value
-    of this dict is again a dict, now of feature names. The values of this dict
-    is a list of feature.
-
-    An example makes this clear. Let's say you load the C. elegans GTF file
-    from Ensemble and make a feature dict:
-
-    >>> worm_features_dict = HTSeq.make_feature_dict(HTSeq.parse_GFF(
-    ...     "test_data/Caenorhabditis_elegans.WS200.55.gtf.gz"))
-
-    (This command may take a few minutes to deal with the 430,000 features
-    in the GTF file. Note that you may need a lot of RAM if you have millions
-    of features.)
-
-    Then, you can simply access, say, exon 0 of gene "F08E10.4" as follows:
-    >>> worm_features_dict[ 'exon' ][ 'F08E10.4' ][ 0 ]
-    <GenomicFeature: exon 'F08E10.4' at V: 17479353 -> 17479001 (strand '-')>
-    """
-
-    res = {}
-    for f in feature_sequence:
-        if f.type not in res:
-            res[f.type] = {}
-        res_ftype = res[f.type]
-        if f.name not in res_ftype:
-            res_ftype[f.name] = [f]
-        else:
-            res_ftype[f.name].append(f)
-    return res
-
 
 #########################
 # GenomicArray
@@ -676,7 +426,6 @@ def pair_SAM_alignments(
         yield; the other 3 alignments of read 2 are ignored.
         '''
 
-
         while len(almnt_list) > 0:
             a1 = almnt_list.pop(0)
             # Find its mate
@@ -715,9 +464,11 @@ def pair_SAM_alignments(
         if almnt.pe_which == "unknown":
             raise ValueError(
                 "Paired-end read found with 'unknown' 'pe_which' status.")
+
         # FIXME: almnt.not_primary_alignment currently means secondary
         if primary_only and (almnt.not_primary_alignment or almnt.supplementary):
             continue
+
         if almnt.read.name == current_name:
             almnt_list.append(almnt)
         else:
@@ -1085,7 +836,7 @@ class BAM_Reader(object):
     def __init__(self, filename, check_sq=True):
         global pysam
         self.filename = filename
-        self.sf = None  # This one is only used by __getitem__
+        self.sf = None
         self.record_no = -1
         self.check_sq = check_sq
         try:
@@ -1096,11 +847,14 @@ class BAM_Reader(object):
             raise
 
     def __iter__(self):
-        sf = pysam.AlignmentFile(self.filename, "r", check_sq=self.check_sq)
+        if self.sf is None:
+            self.sf = pysam.AlignmentFile(self.filename, "r", check_sq=self.check_sq)
         self.record_no = 0
-        for pa in sf:
-            yield SAM_Alignment.from_pysam_AlignedSegment(pa, sf)
+        for pa in self.sf:
+            yield SAM_Alignment.from_pysam_AlignedSegment(pa, self.sf)
             self.record_no += 1
+        self.sf.close()
+        self.sf = None
 
     def fetch(self, reference=None, start=None, end=None, region=None):
         sf = pysam.AlignmentFile(self.filename, "r", check_sq=self.check_sq)
@@ -1131,16 +885,21 @@ class BAM_Reader(object):
                 "Use a HTSeq.GenomicInterval to access regions within .bam-file!")
         if self.sf is None:
             self.sf = pysam.AlignmentFile(self.filename, "r", check_sq=self.check_sq)
-            # NOTE: pysam 0.9 has renames _hasIndex into has_index
-            if (hasattr(self.sf, '_hasIndex') and (not self.sf._hasIndex())) or (not self.sf.has_index()):
-                raise ValueError(
-                    "The .bam-file has no index, random-access is disabled!")
+        if (hasattr(self.sf, '_hasIndex') and (not self.sf._hasIndex())) or (not self.sf.has_index()):
+            raise ValueError(
+                "The .bam-file has no index, random-access is disabled!")
         for pa in self.sf.fetch(iv.chrom, iv.start + 1, iv.end):
             yield SAM_Alignment.from_pysam_AlignedRead(pa, self.sf)
 
     def get_header_dict(self):
-        sf = pysam.AlignmentFile(self.filename, "r", check_sq=self.check_sq)
-        return sf.header
+        if self.sf is None:
+            self.sf = pysam.AlignmentFile(self.filename, "r", check_sq=self.check_sq)
+        return self.sf.header
+
+    def get_template(self):
+        if self.sf is None:
+            self.sf = pysam.AlignmentFile(self.filename, "r", check_sq=self.check_sq)
+        return self.sf
 
 
 # NOTE: this will be deprecated
